@@ -1,14 +1,13 @@
 /*
- *  $Id: dlg_keys.c,v 1.18 2006/01/27 01:43:19 tom Exp $
+ *  $Id: dlg_keys.c,v 1.35 2014/01/12 18:21:52 tom Exp $
  *
  *  dlg_keys.c -- runtime binding support for dialog
  *
- *  Copyright 2005,2006	Thomas E. Dickey
+ *  Copyright 2006-2011,2014 Thomas E. Dickey
  *
  *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU Lesser General Public License as
- *  published by the Free Software Foundation; either version 2.1 of the
- *  License, or (at your option) any later version.
+ *  it under the terms of the GNU Lesser General Public License, version 2.1
+ *  as published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful, but
  *  WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -35,6 +34,7 @@ LIST_BINDINGS {
     DLG_KEYS_BINDING *binding;	/* list of bindings */
 };
 
+#define WILDNAME "*"
 static LIST_BINDINGS *all_bindings;
 static const DLG_KEYS_BINDING end_keys_binding = END_KEYS_BINDING;
 
@@ -53,7 +53,7 @@ dlg_register_window(WINDOW *win, const char *name, DLG_KEYS_BINDING * binding)
 	}
     }
     /* add built-in bindings at the end of the list (see compare_bindings). */
-    if ((p = (LIST_BINDINGS *) calloc(1, sizeof(*p))) != 0) {
+    if ((p = dlg_calloc(LIST_BINDINGS, 1)) != 0) {
 	p->win = win;
 	p->name = name;
 	p->binding = binding;
@@ -62,6 +62,17 @@ dlg_register_window(WINDOW *win, const char *name, DLG_KEYS_BINDING * binding)
 	else
 	    all_bindings = p;
     }
+#if defined(HAVE_DLG_TRACE) && defined(HAVE_RC_FILE)
+    /*
+     * Trace the binding information assigned to this window.  For most widgets
+     * there is only one binding table.  forms have two, so the trace will be
+     * longer.  Since compiled-in bindings are only visible when the widget is
+     * registered, there is no other way to see what bindings are available,
+     * than by running dialog and tracing it.
+     */
+    dlg_trace_msg("# dlg_register_window %s\n", name);
+    dlg_dump_window_keys(dialog_state.trace_output, win);
+#endif
 }
 
 /*
@@ -125,8 +136,8 @@ dlg_register_buttons(WINDOW *win, const char *name, const char **buttons)
 	    continue;
 #endif
 
-	if ((p = (LIST_BINDINGS *) calloc(1, sizeof(*p))) != 0) {
-	    if ((q = (DLG_KEYS_BINDING *) calloc(2, sizeof(*q))) != 0) {
+	if ((p = dlg_calloc(LIST_BINDINGS, 1)) != 0) {
+	    if ((q = dlg_calloc(DLG_KEYS_BINDING, 2)) != 0) {
 		q[0].is_function_key = 0;
 		q[0].curses_key = curses_key;
 		q[0].dialog_key = curses_key;
@@ -190,7 +201,7 @@ int
 dlg_lookup_key(WINDOW *win, int curses_key, int *fkey)
 {
     LIST_BINDINGS *p;
-    int n;
+    DLG_KEYS_BINDING *q;
 
     /*
      * Ignore mouse clicks, since they are already encoded properly.
@@ -209,19 +220,30 @@ dlg_lookup_key(WINDOW *win, int curses_key, int *fkey)
     } else
 #endif
     if (*fkey == 0 || curses_key < KEY_MAX) {
+	const char *name = WILDNAME;
+	if (win != 0) {
+	    for (p = all_bindings; p != 0; p = p->link) {
+		if (p->win == win) {
+		    name = p->name;
+		    break;
+		}
+	    }
+	}
 	for (p = all_bindings; p != 0; p = p->link) {
-	    if (p->win == win || p->win == 0) {
+	    if (p->win == win ||
+		(p->win == 0 &&
+		 (!strcmp(p->name, name) || !strcmp(p->name, WILDNAME)))) {
 		int function_key = (*fkey != 0);
-		for (n = 0; p->binding[n].is_function_key >= 0; ++n) {
+		for (q = p->binding; q->is_function_key >= 0; ++q) {
 		    if (p->buttons
 			&& !function_key
-			&& p->binding[n].curses_key == (int) dlg_toupper(curses_key)) {
+			&& q->curses_key == (int) dlg_toupper(curses_key)) {
 			*fkey = 0;
-			return p->binding[n].dialog_key;
+			return q->dialog_key;
 		    }
-		    if (p->binding[n].curses_key == curses_key
-			&& p->binding[n].is_function_key == function_key) {
-			*fkey = p->binding[n].dialog_key;
+		    if (q->curses_key == curses_key
+			&& q->is_function_key == function_key) {
+			*fkey = q->dialog_key;
 			return *fkey;
 		    }
 		}
@@ -282,7 +304,11 @@ dlg_result_key(int dialog_key, int fkey GCC_UNUSED, int *resultp)
     if (dialog_key == ESC) {
 	*resultp = DLG_EXIT_ESC;
 	done = TRUE;
+    } else if (dialog_key == ERR) {
+	*resultp = DLG_EXIT_ERROR;
+	done = TRUE;
     }
+
     return done;
 }
 
@@ -292,10 +318,18 @@ typedef struct {
     int code;
 } CODENAME;
 
+#define ASCII_NAME(name,code)  { #name, code }
 #define CURSES_NAME(upper) { #upper, KEY_ ## upper }
 #define COUNT_CURSES  sizeof(curses_names)/sizeof(curses_names[0])
 static const CODENAME curses_names[] =
 {
+    ASCII_NAME(ESC, '\033'),
+    ASCII_NAME(CR, '\r'),
+    ASCII_NAME(LF, '\n'),
+    ASCII_NAME(FF, '\f'),
+    ASCII_NAME(TAB, '\t'),
+    ASCII_NAME(DEL, '\177'),
+
     CURSES_NAME(DOWN),
     CURSES_NAME(UP),
     CURSES_NAME(LEFT),
@@ -405,6 +439,10 @@ static const CODENAME dialog_names[] =
     DIALOG_NAME(FIELD_LAST),
     DIALOG_NAME(FIELD_NEXT),
     DIALOG_NAME(FIELD_PREV),
+    DIALOG_NAME(FORM_FIRST),
+    DIALOG_NAME(FORM_LAST),
+    DIALOG_NAME(FORM_NEXT),
+    DIALOG_NAME(FORM_PREV),
     DIALOG_NAME(GRID_UP),
     DIALOG_NAME(GRID_DOWN),
     DIALOG_NAME(GRID_LEFT),
@@ -415,13 +453,15 @@ static const CODENAME dialog_names[] =
     DIALOG_NAME(ENTER),
     DIALOG_NAME(BEGIN),
     DIALOG_NAME(FINAL),
-    DIALOG_NAME(SELECT)
+    DIALOG_NAME(SELECT),
+    DIALOG_NAME(HELPFILE),
+    DIALOG_NAME(TRACE)
 };
 
 static char *
 skip_white(char *s)
 {
-    while (*s != '\0' && isspace(CharOf(*s)))
+    while (*s != '\0' && isspace(UCH(*s)))
 	++s;
     return s;
 }
@@ -429,7 +469,7 @@ skip_white(char *s)
 static char *
 skip_black(char *s)
 {
-    while (*s != '\0' && !isspace(CharOf(*s)))
+    while (*s != '\0' && !isspace(UCH(*s)))
 	++s;
     return s;
 }
@@ -469,9 +509,9 @@ compare_bindings(LIST_BINDINGS * a, LIST_BINDINGS * b)
     if (a->win == b->win) {
 	if (!strcmp(a->name, b->name)) {
 	    result = a->binding[0].curses_key - b->binding[0].curses_key;
-	} else if (!strcmp(b->name, "*")) {
+	} else if (!strcmp(b->name, WILDNAME)) {
 	    result = -1;
-	} else if (!strcmp(a->name, "*")) {
+	} else if (!strcmp(a->name, WILDNAME)) {
 	    result = 1;
 	} else {
 	    result = dlg_strcmp(a->name, b->name);
@@ -493,15 +533,15 @@ compare_bindings(LIST_BINDINGS * a, LIST_BINDINGS * b)
 static DLG_KEYS_BINDING *
 make_binding(char *widget, int curses_key, int is_function, int dialog_key)
 {
-    LIST_BINDINGS *entry;
-    DLG_KEYS_BINDING *data;
+    LIST_BINDINGS *entry = 0;
+    DLG_KEYS_BINDING *data = 0;
     char *name;
     LIST_BINDINGS *p, *q;
     DLG_KEYS_BINDING *result = find_binding(widget, curses_key);
 
     if (result == 0
-	&& (entry = calloc(1, sizeof(LIST_BINDINGS))) != 0
-	&& (data = calloc(2, sizeof(DLG_KEYS_BINDING))) != 0
+	&& (entry = dlg_calloc(LIST_BINDINGS, 1)) != 0
+	&& (data = dlg_calloc(DLG_KEYS_BINDING, 2)) != 0
 	&& (name = dlg_strclone(widget)) != 0) {
 
 	entry->name = name;
@@ -527,6 +567,10 @@ make_binding(char *widget, int curses_key, int is_function, int dialog_key)
 	    entry->link = p;
 	}
 	result = data;
+    } else if (entry != 0) {
+	free(entry);
+	if (data)
+	    free(data);
     }
 
     return result;
@@ -539,7 +583,7 @@ make_binding(char *widget, int curses_key, int is_function, int dialog_key)
  *
  * The curses key "should" be one of the names (ignoring case) from
  * curses_names[], but may also be a single control character (prefix "^" or
- * "~" depending on whether it is C0 or C1), or an escaped single character. 
+ * "~" depending on whether it is C0 or C1), or an escaped single character.
  * Binding a printable character with dialog is possible but not useful.
  *
  * The dialog key must be one of the names from dialog_names[].
@@ -565,6 +609,7 @@ dlg_parse_bindkey(char *params)
     p = skip_black(p);
     if (p != widget && *p != '\0') {
 	*p++ = '\0';
+	p = skip_white(p);
 	q = p;
 	while (*p != '\0' && curses_key < 0) {
 	    if (escaped) {
@@ -586,12 +631,12 @@ dlg_parse_bindkey(char *params)
 		modified = *p;
 	    } else if (*p == '~') {
 		modified = *p;
-	    } else if (isspace(CharOf(*p))) {
+	    } else if (isspace(UCH(*p))) {
 		break;
 	    }
 	    ++p;
 	}
-	if (!isspace(CharOf(*p))) {
+	if (!isspace(UCH(*p))) {
 	    ;
 	} else {
 	    *p++ = '\0';
@@ -606,7 +651,7 @@ dlg_parse_bindkey(char *params)
 		    for (xx = 0; xx < COUNT_CURSES; ++xx) {
 			if (!dlg_strcmp(curses_names[xx].name, q)) {
 			    curses_key = curses_names[xx].code;
-			    is_function = TRUE;
+			    is_function = (curses_key >= KEY_MIN);
 			    break;
 			}
 		    }
@@ -693,30 +738,53 @@ dump_one_binding(FILE *fp, const char *widget, DLG_KEYS_BINDING * binding)
     fputc('\n', fp);
 }
 
+/*
+ * Dump bindings for the given window.  If it is a null, then this dumps the
+ * initial bindings which were loaded from the rc-file that are used as
+ * overall defaults.
+ */
 void
-dlg_dump_keys(FILE *fp)
+dlg_dump_window_keys(FILE *fp, WINDOW *win)
 {
-    LIST_BINDINGS *p;
-    const char *last = "";
-    unsigned n;
-    unsigned count = 0;
+    if (fp != 0) {
+	LIST_BINDINGS *p;
+	DLG_KEYS_BINDING *q;
+	const char *last = "";
 
-    for (p = all_bindings; p != 0; p = p->link) {
-	if (p->win == 0) {
-	    ++count;
-	}
-    }
-    if (count != 0) {
-	for (p = all_bindings, n = 0; p != 0; p = p->link) {
-	    if (p->win == 0) {
+	for (p = all_bindings; p != 0; p = p->link) {
+	    if (p->win == win) {
 		if (dlg_strcmp(last, p->name)) {
 		    fprintf(fp, "\n# key bindings for %s widgets\n",
-			    !strcmp(p->name, "*") ? "all" : p->name);
+			    !strcmp(p->name, WILDNAME) ? "all" : p->name);
 		    last = p->name;
 		}
-		dump_one_binding(fp, p->name, p->binding);
+		for (q = p->binding; q->is_function_key >= 0; ++q) {
+		    dump_one_binding(fp, p->name, q);
+		}
 	    }
 	}
     }
 }
+
+/*
+ * Dump all of the bindings which are not specific to a given widget, i.e.,
+ * the "win" member is null.
+ */
+void
+dlg_dump_keys(FILE *fp)
+{
+    if (fp != 0) {
+	LIST_BINDINGS *p;
+	unsigned count = 0;
+
+	for (p = all_bindings; p != 0; p = p->link) {
+	    if (p->win == 0) {
+		++count;
+	    }
+	}
+	if (count != 0) {
+	    dlg_dump_window_keys(fp, 0);
+	}
+    }
+}
 #endif /* HAVE_RC_FILE */
